import argparse
import fracsuite.scalper
from fracsuite.splinters.analyzerConfig import AnalyzerConfig

mainParser = argparse.ArgumentParser(prog="fracsuite", description="A suite of tools for analyzing impact fragmentation images.")



args = mainParser.parse_args()
